@@ -39,9 +39,9 @@ typedef __u32 u32;
 typedef __u64 u64;
 #include <linux/bsg.h>
 #include <scsi/sg.h>
-#include <scsi/fc/fc_gs.h>
-#include <scsi/fc/fc_ns.h>
-#include <scsi/scsi_bsg_fc.h>
+#include "fc_gs.h"
+#include "fc_ns.h"
+#include "scsi_bsg_fc.h"
 
 static bool quiet = false;
 
@@ -150,7 +150,7 @@ static u16 ct_rjt(u8 reason, u8 explan) {
 }
 
 static u8 ct_rjt_reason(u16 rjt) {
-	return (u8) rjt >> 8;
+	return (u8)(rjt >> 8);
 }
 
 static u8 ct_rjt_explan(u16 rjt) {
